/*
 * Desenvolvido para a disciplina Programacao 1
 * Curso de Bacharelado em Cincia da Computao
 * Departamento de Sistemas e Computao
 * Universidade Federal da Paraba
 *
 * Copyright (C) 1999 Universidade Federal da Paraba.
 * No redistribuir sem permisso.
 */
package p1.aplic.banco;

import java.util.*;
import java.io.*;
import p1.aplic.geral.*;

/**
 * Classe abstrata de conta bancria com implementaes default de alguns mtodos.
 *
 * @author   Jacques Philippe Sauv, jacques@dsc.ufpb.br
 * @version 1.1
 * <br>
 * Copyright (C) 1999 Universidade Federal da Paraba.
 */
public abstract class Conta implements Serializable {
  static final long serialVersionUID = 8972898734498390810L;
	private int			   nmero;
	private Pessoa     titular;
	private Data       dataAbertura;
	private Moeda		   saldo;
	private Movimento  movimento;

	// construtores
	/**
	 * Cria uma conta a partir de uma pessoa e nmero de conta.
	 * @param titular O titular da conta.
	 * @param nmero O nmero da conta.
	 */
	public Conta(Pessoa titular, int nmero) {
		this.nmero = nmero;
		this.titular = titular;
		dataAbertura = new Data();	// A data representa este instante de tempo
		saldo = new Real(0.0);		// Conta em reais e zerada
		movimento = new Movimento();
	}

	// Mtodos
	/**
	 * Recupera o nmero da conta.
	 * @return O nmero da conta.
	 */
	public int getNmero() {
		return nmero;
	}

	/**
	 * Recupera o titular da conta.
	 * @return O titular da conta.
	 */
	public Pessoa getTitular() {
		return titular;
	}

	/**
	 * Recupera o nome do titular da conta.
	 * @return O nome do titular da conta.
	 */
	/* feito para ajudar os principiantes escondendo a classe Pessoa no inicio */
	public String getNome() {
		return titular.getNome();
	}

	/**
	 * Recupera o CPF do titular da conta.
	 * @return O CPF do titular da conta.
	 */
	/* feito para ajudar os principiantes escondendo a classe Pessoa no inicio */
	public String getCPF() {
		return titular.getCPF();
	}

	/**
	 * Recupera a data de abertura da conta.
	 * @return A data de abertura da conta.
	 */
	public Data getDataAbertura() {
		return dataAbertura;
	}
   
	/**
	 * Recupera o saldo da conta.
	 * @return O saldo da conta como double.
	 */
	public double getSaldo() {
		return saldo.getValor();
	}

	/**
	 * Recupera o saldo da conta.
	 * @return O saldo da conta como Moeda.
	 */
	public Moeda getSaldoMonetrio() {
		return saldo;
	}

	/**
	 * Recupera o movimento da conta.
	 * @return O movimento da conta.
	 */
	public Movimento getMovimento() {
		return movimento;
	}

	/**
	 * Fornece um Iterator para varrer as transaes por data.
	 * @return O Iterator.
	 */
	public Iterator getTransaes() {
		return getMovimento().getTransaes();
	}

	/**
	 * Transfere um valor para outra conta.
	 * @param destino A conta destino.
	 * @param valor O valor a transferir.
	 * @return O sucesso ou no da operao.
	 */
	public boolean transferir(Conta destino, double valor) {
		return transferir(destino, valor, "transferencia para conta " + destino.getNmero());
	}

	/**
	 * Transfere um valor para outra conta.
	 * @param destino A conta destino.
	 * @param valor O valor a transferir.
	 * @param descr A descrio da transao.
	 * @return O sucesso ou no da operao. H insucesso se no houver saldo suficiente.
	 */
	public boolean transferir(Conta destino, double valor, String descr) {
		if(saldo.getValor() - valor >= 0) {
			this.receber(-valor);
			destino.receber(valor);
			logarTransacao(this, destino, new Real(valor), descr);
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Recebe valor na conta.
	 * @param valor O valor a receber.
	 */
    // observe que o mtodo *no*  pblico
	void receber(double valor) {
		saldo.setValor(saldo.getValor() + valor);
	}

	/**
	 * Efetua um depsito numa conta.
	 * @param valor O valor a depositar.
	 */
	public boolean depositar(double valor) {
		// credita a conta e debita o caixa
		return Agencia.getContaCaixa().transferir(this, valor, "deposito");
	}

	/**
	 * Efetua sacada na conta.
	 * @param valor O valor a sacar.
	 * @return O sucesso ou no da operao.
	 */
	public boolean sacar(double valor) {
		// debita a conta e credita o caixa
		return transferir(Agencia.getContaCaixa(), valor, "saque");
	}

	/**
	 * Efetua log de uma transao no movimento.
     * @param contaDbito A conta que foi debitada.
     * @param contaCrdito A conta que foi creditada.
	 * @param valor O valor da transao.
     * @param descr A descrio da transao.
	 */
    // observe que p mtodo *no*  pblico
    void logarTransacao(Conta contaDbito, Conta contaCrdito, Moeda valor, String descr) {
		Transacao transacao = new Transacao(new Data(), contaDbito, contaCrdito, valor, descr);
		contaDbito.getMovimento().add(transacao);
		contaCrdito.getMovimento().add(transacao);
		Agencia.getMovimento().add(transacao);
	}

	/**
	 * Transforma os dados da conta em um String.
	 * @return O string com os dados da conta.
	 */
	public String toString() {
		return "numero " + getNmero()
				+ ", titular " + getNome()
				+ ", data " + getDataAbertura().DDMMAAAA()
				+ ", saldo " + getSaldoMonetrio();
	}

	/**
	 * Fecha a conta.
	 * @throws NaoPodeFecharContaException Quando se tenta fechar uma conta com saldo no zero.
	 */
	public void fechar() throws NaoPodeFecharContaException  {
		Agencia.fecharConta(nmero);
	}

	/**
	 * Cria um extrato com todas as transaes entre duas datas.
	 * @param dataInicial A data inicial do extrato (inclusive).
	 * @param dataFinal A data final do extrato (inclusive).
	 */
	public Extrato criarExtrato(Data dataInicial, Data dataFinal) {
		return new Extrato(dataInicial, dataFinal, movimento);
	}
}
