/*
 * Desenvolvido para a disciplina Programacao 2
 * Curso de Bacharelado em Cincia da Computao
 * Departamento de Sistemas e Computao
 * Universidade Federal da Paraba
 *
 * Copyright (C) 2001 Universidade Federal da Paraba.
 * No redistribuir sem permisso.
 */
package p1.aplic.mancala.jogo;
import java.util.*;

/**
 * Um tabuleiro do jogo de Mancala com 6 buracos e uma Kalaha para cada jogador.
 *
 * @author   Jacques Philippe Sauv, jacques@dsc.ufpb.br
 * @version 1.0
 * <br>
 * Copyright (C) 2001 Universidade Federal da Paraba.
 */
public class Tabuleiro {
  /**
   * Posio "em baixo" do tabuleiro, com buracos 1 a 6 e Kalaha  direita.
   */
  public final static int EM_BAIXO = 0;

  /**
   * Posio "em cima" do tabuleiro, com buracos 8 a 13 e Kalaha  esquerda.
   */
  public final static int EM_CIMA = 1;

  /**
   * Posio de um buraco indicando uma Kalaha.
   */
  public final static int KALAHA = 2;

  /**
   * Nmero de buracos por jogador (sem incluir Kalahas)
   */
  public final static int NUM_BURACOS = 6;

  private final static int TOTAL_BURACOS = 2*(NUM_BURACOS+1);
  private final static int PRIMEIRO_BURACO_EM_BAIXO = 0;
  private final static int LTIMO_BURACO_EM_BAIXO = NUM_BURACOS-1;
  private final static int KALAHA_DIREITA = NUM_BURACOS;
  private final static int PRIMEIRO_BURACO_EM_CIMA = NUM_BURACOS+1;
  private final static int LTIMO_BURACO_EM_CIMA = TOTAL_BURACOS-2;
  private final static int KALAHA_ESQUERDA = TOTAL_BURACOS-1;
  private Buraco[] buracos;

  /**
   * Construtor de um tabuleiro padro vazio para jogar Mancala.
   */
  Tabuleiro() {
    buracos = new Buraco[TOTAL_BURACOS];
    for(int i = 0; i < TOTAL_BURACOS; i++) {
      buracos[i] = new Buraco(i, 0);
    }
  }

  /**
   * Informa a Kalaha correspondendo a uma posio (EM_BAIXO ou EM_CIMA)
   * A Kalaha do jogador "em baixo" est  direita.
   * A Kalaha do jogador "em cima" est  esquerda.
   *
   * @param   posio   A posio de interesse (EM_BAIXO ou EM_CIMA).
   * @return  O buraco correpondendo  Kalaha da posio de interesse.
   */
  Buraco getKalaha(int posio) {
    return posio == EM_BAIXO ?
           buracos[KALAHA_DIREITA] :
           buracos[KALAHA_ESQUERDA];
  }

  /**
   * Informa o buraco correspondendo a um nmero de buraco
   * de uma certa posio (EM_BAIXO ou EM_CIMA).
   *
   * @param   posio   A posio de interesse (EM_BAIXO ou EM_CIMA).
   * @param   nmeroDoBuraco   O nmero do buraco (0 a 5) de interesse.
   * @return  O buraco de interesse.
   */
  public Buraco getBuraco(int posio, int nmeroDoBuraco) {
    return posio == EM_CIMA ? buracos[PRIMEIRO_BURACO_EM_CIMA + nmeroDoBuraco] :
                                buracos[PRIMEIRO_BURACO_EM_BAIXO + nmeroDoBuraco];
  }

  /**
   * Verifica se um lado do tabuleiro est vazio.
   *
   * @param   posio   A posio de interesse (EM_BAIXO ou EM_CIMA).
   * @return  true se o lado do tabuleiro indicado pela posio est vazio,
   *          isto , com os 6 buracos vazios.
   */
  boolean ladoEstVazio(int posio) {
    for(int i = 0; i < Tabuleiro.NUM_BURACOS; i++) {
      Buraco buraco = getBuraco(posio, i);
      if(buraco.getNmeroDeSementes() > 0) {
        return false;
      }
    }
    return true;
  }

  /**
   * Informa o prximo buraco depois de um certo buraco, incluindo Kalahas,
   * no sentido do jogo (anti-horrio).
   *
   * @param   buraco   O buraco de referncia.
   * @return  O buraco depois do buraco de referncia.
   */
  public Buraco prximoBuraco(Buraco buraco) {
    return buracos[(buraco.getNmero()+1)%TOTAL_BURACOS];
  }

  /**
   * Dado um buraco de referncia, informa o buraco do lado oposto do tabuleiro.
   *
   * @param   buraco   O buraco de referncia.
   * @return  O buraco do lado oposto do tabuleiro
   */
  Buraco buracoOposto(Buraco buraco) {
    int index = getPosio(buraco) == EM_BAIXO ?
                KALAHA_DIREITA + (KALAHA_DIREITA - buraco.getNmero()):
                KALAHA_ESQUERDA + (KALAHA_ESQUERDA - buraco.getNmero());
    return buracos[index % TOTAL_BURACOS];
  }

  /**
   * Informa se um buraco est EM_CIMA, EM_BAIXO, ou  KALAHA.
   *
   * @param   buraco   O buraco de interesse.
   * @return  EM_BAIXO, para buracos de cima
   *          EM_CIMA, para buracos de baixo
   *          KALAHA para as duas Kalahas.
   */
  int getPosio(Buraco buraco) {
    int num = buraco.getNmero();
    if(PRIMEIRO_BURACO_EM_BAIXO <= num && num <= LTIMO_BURACO_EM_BAIXO) {
      return EM_BAIXO;
    } else if(PRIMEIRO_BURACO_EM_CIMA <= num && num <= LTIMO_BURACO_EM_CIMA) {
      return EM_CIMA;
    } else {
      return KALAHA;
    }
  }

  /**
   * Converte um nmero absoluto (de 0 a 13) para um ndice de buraco (0 a 5)
   * Se o buraco no corresponder  posio desejada, lana exceo.
   * (Isso fede e precisa de refatoramento)
   *
   * @param   numAbsoluto   Um nmero de buraco do tabuleiro, de 0 a 13.
   * @param   posio   A posio (EM_BAIXO, EM_CIMA) na qual este buraco deve estar.
   * @return  O ndice do buraco correspondente (de 0 a 5).
   * @throws  MancalaException se O buraco no se localiza na posio indicada.
   */
  public int numAIndex(int numAbsoluto, int posio) throws MancalaException {
    if(posio == EM_CIMA) {
      if(PRIMEIRO_BURACO_EM_CIMA <= numAbsoluto && numAbsoluto <= LTIMO_BURACO_EM_CIMA) {
        return numAbsoluto - PRIMEIRO_BURACO_EM_CIMA;
      }
    } else {
      if(PRIMEIRO_BURACO_EM_BAIXO <= numAbsoluto && numAbsoluto <= LTIMO_BURACO_EM_BAIXO) {
        return numAbsoluto - PRIMEIRO_BURACO_EM_BAIXO;
      }
    }
    throw new MancalaException("Esse buraco nao eh seu!");
  }

  /**
   * Duplica o tabuleiro.  um Deep Copy.
   *
   * @return  O clone do objeto.
   */
  public Object clone() {
    Tabuleiro novo = new Tabuleiro();
    for(int i = 0; i < TOTAL_BURACOS; i++) {
      novo.buracos[i].adicionaSementes(this.buracos[i].getNmeroDeSementes());
    }
    return novo;
  }
}

